/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2, with the OpenSSL exception described
 * in the README file. */

/*!
 * \file filterplugin.cpp
 *
 * \brief FilterPlugin class defintion.
 */

#include "imspector.h"

/*!
 * \brief Constructor. Clears the handle.
 */
FilterPlugin::FilterPlugin()
{
	handle = NULL;
}

/*!
 * \brief Destructor. Does nothing at all.
 */
FilterPlugin::~FilterPlugin()
{
}

/*!
 * \brief Loads a plugin.
 *
 * Uses dlopen to load a shared object and sets up the function pointers for
 * the plugin.
 *
 * \param filename The filename of the shared object to load.
 * \return false for failure.
 */
bool FilterPlugin::loadplugin(std::string filename)
{
	handle = dlopen(filename.c_str(), RTLD_LAZY);
	
	if (!handle)
	{
		syslog(LOG_ERR, "dlopen(): %s\n", dlerror());
		return false;
	}
	
	initfilterplugin = (initfilterplugintype) dlsym(handle, "initfilterplugin");
	closefilterplugin = (closefilterplugintype) dlsym(handle, "closefilterplugin");
	filter = (filtertype) dlsym(handle, "filter");
	
	if (!initfilterplugin || !closefilterplugin || !filter)
	{
		syslog(LOG_ERR, "%s: dlsym(): %s\n", filename.c_str(), dlerror());
		return false;
	}
	
	return true;
}

/*!
 * \brief Unloads a plugin using dlclose.
 *
 * \return Always returns true.
 */
bool FilterPlugin::unloadplugin(void)
{
	if (handle) dlclose(handle);
	
	return true;
}

/*!
 * \brief Wrapper for the init plugin function.
 *
 * \param options The options hash.
 * \param debugmode The debug flag.
 * \return Whatever the underlying init function returns.
 */
bool FilterPlugin::callinitfilterplugin(class Options &options, bool debugmode)
{
	return (*initfilterplugin)(filterplugininfo, options, debugmode);
}

/*!
 * \brief Wrapper for the close plugin function.
 */
void FilterPlugin::callclosefilterplugin(void)
{
	return (*closefilterplugin)();
}

/*!
 * \brief Wrapper for the filter plugin function.
 *
 * This plugin function does the actual filtering.
 *
 * \param originalbuffer The buffer before it was mofidified by any filter plugin.
 * \param modifiedbuffer The buffer after it was modified.
 * \param imevent The list of imevents to filter.
 */
bool FilterPlugin::callfilter(char *originalbuffer, char *modifiedbuffer, struct imevent &imevent)
{
	return (*filter)(originalbuffer, modifiedbuffer, imevent);
}
