/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2, with the OpenSSL exception described
 * in the README file. */

/*!
 * \file loggingplugin.cpp
 *
 * \brief LoggingPlugin class definition.
 */

#include "imspector.h"

/*!
 * \brief Constructor. Clears the handle.
 */
LoggingPlugin::LoggingPlugin()
{
	handle = NULL;
}

/*!
 * \brief Destructor. Does nothing at all.
 */
LoggingPlugin::~LoggingPlugin()
{
}

/*!
 * \brief Loads a plugin.
 *
 * Uses dlopen to load a shared object and sets up the function pointers for
 * the plugin.
 *
 * \param filename The filename of the shared object to load.
 * \return false for failure.
 */
bool LoggingPlugin::loadplugin(std::string filename)
{
	handle = dlopen(filename.c_str(), RTLD_LAZY);
	
	if (!handle)
	{
		syslog(LOG_ERR, "dlopen(): %s\n", dlerror());
		return false;
	}
	
	initloggingplugin = (initloggingplugintype) dlsym(handle, "initloggingplugin");
	closeloggingplugin = (closeloggingplugintype) dlsym(handle, "closeloggingplugin");
	logevents = (logeventstype) dlsym(handle, "logevents");
	
	if (!initloggingplugin || !closeloggingplugin || !logevents)
	{
		syslog(LOG_ERR, "%s: dlsym(): %s\n", filename.c_str(), dlerror());
		return false;
	}
	
	return true;
}

/*!
 * \brief Unloads a plugin using dlclose.
 *
 * \return Always returns true.
 */
bool LoggingPlugin::unloadplugin(void)
{
	if (handle) dlclose(handle);
	
	return true;
}

/*!
 * \brief Wrapper for the init plugin function.
 *
 * \param options The options hash.
 * \param debugmode The debug flag.
 * \return Whatever the underlying init function returns.
 */
bool LoggingPlugin::callinitloggingplugin(class Options &options, bool debugmode)
{
	return (*initloggingplugin)(loggingplugininfo, options, debugmode);
}

/*!
 * \brief Wrapper for the close plugin function.
 */
void LoggingPlugin::callcloseloggingplugin(void)
{
	return (*closeloggingplugin)();
}

/*!
 * \brief Wrapper to the logevents function.
 *
 * This plugin function does the actual logging.
 *
 * \param imevents The list of events to send to the logging plugin for writing
 * to disk or database, etc.
 * \return Whatever the underlying plugin returns, but generally 0 for success.
 */
int LoggingPlugin::calllogevents(std::vector<struct imevent> &imevents)
{
	return (*logevents)(imevents);
}
