/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2, with the OpenSSL exception described
 * in the README file. */

/*!
 * \file protocolplugin.cpp
 *
 * \brief ProtocolPlugin class definition.
 */

#include "imspector.h"

/*!
 * \brief Constructor. Clears the handle.
 */
ProtocolPlugin::ProtocolPlugin()
{
	handle = NULL;
}

/*!
 * \brief Destructor. Does nothing at all.
 */
ProtocolPlugin::~ProtocolPlugin()
{
}

/*!
 * \brief Loads a plugin.
 *
 * Uses dlopen to load a shared object and sets up the function pointers for
 * the plugin.
 *
 * \param filename The filename of the shared object to load.
 * \return false for failure.
 */
bool ProtocolPlugin::loadplugin(std::string filename)
{
	handle = dlopen(filename.c_str(), RTLD_LAZY);
	
	if (!handle)
	{
		syslog(LOG_ERR, "dlopen(): %s\n", dlerror());
		return false;
	}
	
	initprotocolplugin = (initprotocolplugintype) dlsym(handle, "initprotocolplugin");
	closeprotocolplugin = (closeprotocolplugintype) dlsym(handle, "closeprotocolplugin");
	processpacket = (processpackettype) dlsym(handle, "processpacket");
	generatemessagepacket = (generatemessagepackettype) dlsym(handle, "generatemessagepacket");
	
	if (!initprotocolplugin || !closeprotocolplugin || !processpacket)
	{
		syslog(LOG_ERR, "%s: dlsym(): %s\n", filename.c_str(), dlerror());
		return false;
	}
	
	return true;
}

/*!
 * \brief Unloads a plugin using dlclose.
 *
 * \return Always returns true.
 */
bool ProtocolPlugin::unloadplugin(void)
{
	if (handle) dlclose(handle);
	
	return true;
}

/*!
 * \brief Wrapper for the init plugin function.
 *
 * \param options The options hash.
 * \param debugmode The debug flag.
 * \return Whatever the underlying init function returns.
 */
bool ProtocolPlugin::callinitprotocolplugin(class Options &options, bool debugmode)
{
	return (*initprotocolplugin)(protocolplugininfo, options, debugmode);
}

/*!
 * \brief Wrapper for the close plugin function.
 */
void ProtocolPlugin::callcloseprotocolplugin(void)
{
	return (*closeprotocolplugin)();
}

/*!
 * \brief Wrapper to the processpacket function.
 *
 * This plugin functions job is key. It should pull packets from the incomingsock
 * and decode them, putting an imevent onto each event as it finds them. Another
 * key thing it has to do is copy the pulled down data into the replybuffer, so
 * the main connnection handling code and send the packet on to the intended
 * recipient; which is either the IM server or the IM client.
 *
 * A protocol handler can use any Socket method it likes to pull the data down,
 * depending on the underlying protocol. Line based protocols (like IRC) would be
 * implemented using Socket::getline, but other options are available.
 *
 * \param outgoing Wether this packet is an "outgoing" one or not.
 * \param incomingsock The socket to pull packets out of.
 * \param replybuffer Where to put packets that we get from the socket.
 * \param replybufferlength How many packets we pulled from the socket.
 * \param imevents The list of events we have found in this packet. Events are pushed
 * onto this list as we decode them.
 * \param clientaddress The address of the client that is responsible for this
 * connection, incase we need it.
 * \return Whatever the underlying plugin returns, but generally 0 for success.
 */
int ProtocolPlugin::callprocesspacket(bool outgoing, class Socket &incomingsock, char *replybuffer,
	int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress)
{
	return (*processpacket)(outgoing, incomingsock, replybuffer,
		replybufferlength, imevents, clientaddress);
}

/*!
 * \brief Wrapper for the generatemessagepacket function.
 *
 * This function is responsible for formatting a message packet for the particular
 * protocol.  Unlike processpacket, a protocol plugin does not have to
 * implement this function.
 *
 * \param response A response structure, which has various bits of info describing
 * the message we want to generate.
 * \param replybuffer Where to shove the protocol-formatted packet.
 * \param replybufferlength And the size of that packet.
 * \return Whatever the underlying plugin returns, but generally 0 for success.
 * Can also return 1 if the protocol handler does not implement this function.
 */
int ProtocolPlugin::callgeneratemessagepacket(struct response &response, char *replybuffer,
	int *replybufferlength)
{
	if (generatemessagepacket)
		return (*generatemessagepacket)(response, replybuffer, replybufferlength);
	else
		return 1;
}
