/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2, with the OpenSSL exception described
 * in the README file. */

/*!
 * \file protocolplugin.h
 *
 * \brief ProtocolPlugin class.
 */

/*!
 * \brief Info structure for a protocol plugin.
 *
 * For protocol plugins, this includes a name, which port the protocol runs on,
 * and, for SSL protoocols, which port is the native SSL version. Also special
 * for protocol plugins, there is a short-form protocolname field.
 */
struct protocolplugininfo
{
	std::string pluginname;   //!< The name of the plugin.
	std::string protocolname; //!< The name of the protocol the plugin handles.
	uint16_t port;            //!< The port the protocol uses, network byte order.
#ifdef HAVE_SSL
	uint16_t sslport;          //!< The port the protocol uses in SSL.
#endif
};

typedef bool (*initprotocolplugintype)(struct protocolplugininfo &protocolplugininfo,
	class Options &options, bool debugmode);
typedef void (*closeprotocolplugintype)(void);
typedef int (*processpackettype)(bool outgoing, class Socket &incomingsock, char *replybuffer, 
	int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress);
typedef int (*generatemessagepackettype)(struct response &response, char *replybuffer, int *replybufferlength);
	
/*!
 * \brief A ProtocolPlugin class encapsulates a protocol plugin shared object file.
 *
 * Protocol plugins, as per all other plugin types, are two things: an object
 * of the class LoggingPlugin, and a shared object wrappered by the object. These
 * objects are usually stored on a vector, so that (say) all the protocol plugins 
 * are tested to see if the outgoing connection is for one of the handled
 * protocols.
 *
 * All plugin types have a plugininfo struct, that describes the plugin.
 */
class ProtocolPlugin
{
	public:
		/*! \brief This is public so we don't have to mess with accessors. */
		struct protocolplugininfo protocolplugininfo;
	
		ProtocolPlugin();
		~ProtocolPlugin();
		bool loadplugin(std::string filename);
		bool unloadplugin(void);
		
		bool callinitprotocolplugin(class Options &options, bool debugmode);
		void callcloseprotocolplugin(void);
		int callprocesspacket(bool outgoing, class Socket &incomingsock, char *replybuffer,
			int *replybufferlength, std::vector<struct imevent> &imevents, std::string &clientaddress);
		int callgeneratemessagepacket(struct response &response, char *replybuffer, int *replybufferlength);

	private:
		void *handle; //!< Our dlopen handle.

		initprotocolplugintype initprotocolplugin;       //!< The initplugin funct pointer.
		closeprotocolplugintype closeprotocolplugin;     //!< The closeplugin funct pointer.
		processpackettype processpacket;                 //!< The processpacket funct pointer.
		generatemessagepackettype generatemessagepacket; //!< The generatemessagepacket funct pointer.
};
