/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 *          
 * Released under the GPL v2, with the OpenSSL exception described
 * in the README file. */

/*!
 * \file responderplugin.cpp
 *
 * \brief ResponderPlugin class defintion.
 */

#include "imspector.h"

/*!
 * \brief Constructor. Clears the handle.
 */
ResponderPlugin::ResponderPlugin()
{
	handle = NULL;
}

/*!
 * \brief Destructor. Does nothing at all.
 */
ResponderPlugin::~ResponderPlugin()
{
}

/*!
 * \brief Loads a plugin.
 *
 * Uses dlopen to load a shared object and sets up the function pointers for
 * the plugin.
 *
 * \param filename The filename of the shared object to load.
 * \return false for failure.
 */
bool ResponderPlugin::loadplugin(std::string filename)
{
	handle = dlopen(filename.c_str(), RTLD_LAZY);
	
	if (!handle)
	{
		syslog(LOG_ERR, "dlopen(): %s\n", dlerror());
		return false;
	}
	
	initresponderplugin = (initresponderplugintype) dlsym(handle, "initresponderplugin");
	closeresponderplugin = (closeresponderplugintype) dlsym(handle, "closeresponderplugin");
	generateresponses = (generateresponsestype) dlsym(handle, "generateresponses");
	
	if (!initresponderplugin || !closeresponderplugin || !generateresponses)
	{
		syslog(LOG_ERR, "%s: dlsym(): %s\n", filename.c_str(), dlerror());
		return false;
	}
	
	return true;
}

/*!
 * \brief Unloads a plugin using dlclose.
 *
 * \return Always returns true.
 */
bool ResponderPlugin::unloadplugin(void)
{
	if (handle) dlclose(handle);
	
	return true;
}

/*!
 * \brief Wrapper for the init plugin function.
 *
 * \param options The options hash.
 * \param debugmode The debug flag.
 * \return Whatever the underlying init function returns.
 */
bool ResponderPlugin::callinitresponderplugin(class Options &options, bool debugmode)
{
	return (*initresponderplugin)(responderplugininfo, options, debugmode);
}

/*!
 * \brief Wrapper for the close plugin function.
 */
void ResponderPlugin::callcloseresponderplugin(void)
{
	return (*closeresponderplugin)();
}

/*!
 * \brief Wrapper for the generateresponses plugin function.
 *
 * This plugin function does the actual generating of responses, based on the list
 * of imevents it sees.
 *
 * \param imevents The list of imevents to send to the responder to see if it
 * would like to genarate a response.
 * \param responses The plugin should fill this vector with a list of responses it
 * would like to see sent.
 * \return The plugin return code, generally 0 for success.
 */
int ResponderPlugin::callgenerateresponses(std::vector<struct imevent> &imevents,
	std::vector<struct response> &responses)
{
	return (*generateresponses)(imevents, responses);
}
