/* IMSpector - Instant Messenger Transparent Proxy Service
 * http://www.imspector.org/
 * (c) Lawrence Manning <lawrence@aslak.net>, 2006
 * 
 * Contributions from:
 *     Ryan Wagoner <ryan@wgnrs.dynu.com>, 2006
 *          
 * Released under the GPL v2, with the OpenSSL exception described
 * in the README file. */

/*!
 * \file tools.cpp
 *
 * \brief Misc handy functions.
 *
 * Various useful things.
 */

#include "imspector.h"

static uint8_t decodebase64char(char c);

/*!
 * \brief Remove the first carriagereturn or linefeed in a string and truncates
 * the string at this point.
 *
 * This is used anywhere that we do not care about line endings.
 *
 * \param buffer The string to strip.
 */
void stripnewline(char *buffer)
{
	char *t;
	
	t = strchr(buffer, '\r'); if (t) *t = '\0';
	t = strchr(buffer, '\n'); if (t) *t = '\0';
}

/*!
 * \brief snprintf for std::string.
 *
 * There is a max size of BUFFER_SIZE, above which the string will get
 * truncated.
 *
 * \param fmt A printf style string.
 * \return The resultant string
 *
 * \note A future version of this string could dynamically resize the temporary
 * buffer.
 */
std::string stringprintf(const char *fmt, ...)
{
	char buffer[BUFFER_SIZE];
	va_list argp;

	memset(buffer, 0, BUFFER_SIZE);
	
	va_start(argp, fmt);
	vsnprintf(buffer, BUFFER_SIZE - 1, fmt, argp);
	va_end(argp);

	return (std::string)(buffer);
}

/*!
 * \brief Tokenise a commandline-like line.
 *
 * Chops the heading first line into its command and arg list. A comamnd line
 * looks like this:
 *
 * COMMAND [ARG1...ARGN]
 *
 * Arguments are optional. This function is used in places like the MSN handler
 * and some other places where we send "commands" with "arguments". The buffer
 * processing will stop when the buffer is exhausted, or when we find a \\r or
 * \\n.
 *
 * \param buffer The input buffer
 * \param command The first word on the "command line".
 * \param args A vector of strings for all the subsequent words.
 * \param argc The count of the number of arguments found.
 * \return A pointer to the end of the buffer.
 */
char *chopline(char* buffer, std::string &command, std::vector<std::string> &args, int &argc)
{
	char *s = buffer;
	
	/* Copy the command, while the character looks valid. */
	for (; *s && *s != ' ' && *s != '\r' && *s != '\n'; s++)
		command.push_back(*s);
	s++;
	
	/* Same for the argument list. */
	argc = 0;
	while (*s && *s != '\r' && *s != '\n')
	{
		std::string arg;
		for (; *s && *s != ' ' && *s != '\r' && *s != '\n'; s++)
			arg.push_back(*s);
		s++;
		
		args.push_back(arg); argc++;
	}
	
	/* Need to advance s to the start of the next line, skipping over the eol chars. */
	for (; *s && (*s == '\r' || *s == '\n'); s++);
	
	return s;
}

/*!
 * \brief Debug handling print function.
 *
 * Prints debug. This is a simple wrapper around syslog to prevent unecessary
 * calls when not in debug mode.
 *
 * \param debugflag if not true, this function does nothing.
 * \param fmt A printf-syle format vararg.
 */
void debugprint(bool debugflag, const char *fmt, ...)
{
	if (!debugflag) return;

	va_list argp;

	va_start(argp, fmt);
	vsyslog(LOG_DEBUG, fmt, argp);
	va_end(argp);
}

/*!
 * \brief Turn base64 encoded text into a binary buffer.
 *
 * At various places in IMSpector we can decode base64 text into the clear
 * text byte representation of the same. This function can do that.
 *
 * \param line The input, in string format.
 * \param buffer The output, in byte format.
 * \param bufferlen The size of the output buffer.
 * \return The number of bytes converted.
 */
int decodebase64(std::string line, uint8_t *buffer, int bufferlen)
{
	uint32_t quartet = 0;
	uint8_t d;
	int c = 0; int len = line.length() - 4;
	
	for (int i = 0; i < len && c < bufferlen - 3; i += 4)
	{
		quartet = 0;
		d = decodebase64char(line[i + 0]);
		quartet = quartet | d;
		d = decodebase64char(line[i + 1]);
		quartet = (quartet << 6) | d;
		d = decodebase64char(line[i + 2]);
		quartet = (quartet << 6) | d;
		d = decodebase64char(line[i + 3]);
		quartet = (quartet << 6) | d;
		d = (quartet & 0xFF0000) >> 16;
		buffer[c++] = (uint8_t) d;
		d = (quartet & 0xFF00) >> 8;
		buffer[c++] = (uint8_t) d;
		d = quartet & 0xFF;
		buffer[c++] = (uint8_t) d;
	}
	
	return c;
}

/*!
 * \brief Decode a single character into base64 byte value.
 *
 * Helper function used by decodebase64.
 *
 * \param c The character in base64 to decode.
 * \return The byte value.
 */
static uint8_t decodebase64char(char c)
{
	uint8_t i = '\0';
	
	switch (c)
	{
		case '+':
			i = 62;
			break;
			
		case '/':
			i = 63;
			break;
			
		case '=':
			i = 0;
			break;
			
		default:
			/* Must be alphanumeric. */
			i = '9' - c;
			if (i > 0x3F)
			{
				/* It is under 9 */
				i = 'Z' - c;
				if (i > 0x3F)
				{
					/* It is after Z. */
					i = 'z' - c;
					if (i > 0x3F)
						i = 0x80;
					else
						/* It is a-z. */
						i = c - 71;
				}
				else
					/* A-Z */
					i = c - 65;
			}
			else 
				i = c + 4;
				
			break;
	}
	
	return i;
}

/*!
 * \brief Packet dumping function.
 *
 * This little function is used to generate trace packets.
 *
 * \param protocol Used to build the final filename for the packet.
 * \param packetcount Also used to buid the filename.
 * \param buffer What to put in the packet file.
 * \param bufferlength The size of the packet.
 *
 * \note No error checking is done; if the packet can't be written then we
 * just carry on.
 */
void tracepacket(const char *protocol, int packetcount, char *buffer, int bufferlength)
{
	std::string filename = stringprintf(TRACE_DIR "/%s.%d.%d", protocol,
		getpid(), packetcount);

	int fd = -1;
	if ((fd = creat(filename.c_str(), 0600)) > 0)
	{
		write(fd, buffer, bufferlength);
		close(fd);
	}
}

/*!
 * \brief DJBs hash function.
 *
 * This is a trivial hash (as in VERY trivial). It takes a null terminated
 * string and returns a 32bit pseudo hash.
 *
 * \param str null terminated string.
 * \return The hash value.
 */
unsigned long hash(const char *str)
{
	unsigned long hash = 5381;
	int c;
	
	while ((c = *str++))
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
	
	return hash;
}

/*!
 * \brief Our XML tag parser.
 *
 * This is an amazingly simple, and somewhat stupid XML parser. It's
 * probably not standards compliant but at least it is simple. It will parse
 * the text, the tag itself, along with it's key=value pairs. It will also
 * indicate wether the tag is a closing tag (ends />). It is definetly the
 * simplest parser possible; it doesn't even have a concept of tag depth or
 * anything.
 *
 * Parsing ends at the end of the closing \> in the tag. Whitespace chars
 * are roughly: space, cr and lf.
 *
 * \param debugmode Used to supress or show debug output.
 * \param buffer The buffer containing the XML tag.
 * \param payload The text before the tag (ie the bit before the opening \<)
 * \param payloadlength The length of the payload.
 * \param tag The actual tag text (... \<tag ...\>)
 * \param closing Wether or not the tag ends in a /.
 * \param params A map containing all the tag options (... \<... key=value\>)
 * \return A pointer to the end of the buffer.
 */
char *parsexmltag(bool debugmode, char* buffer, std::string &payload, int &payloadlength, 
	std::string &tag, bool &closing, std::map<std::string, std::string> &params)
{
	closing = false;

	char *s = buffer;

	/* Copy the payload, while the character looks valid. */
	for (; *s && *s != '<'; s++)
		payload.push_back(*s);
	
	removenewlines(payload);
	
	payloadlength = s - buffer;
	
	debugprint(debugmode, "XML Parse: Payload: %s", payload.c_str());
	
	/* Skip the <. */
	if (*s) s++;
	
	/* Copy the command, while the character looks valid. */
	for (; *s && *s != ' ' && *s != '>'; s++)
		tag.push_back(*s);
		
	removenewlines(tag);

	debugprint(debugmode, "XML Parse: Tag: %s", tag.c_str());

	/* Skip the spaces. */
	for (; *s && (*s == ' ' || *s == '\n' || *s == '\r'); s++);

	while (*s && *s != '>' && *s != '/')
	{
		std::string key; std::string value;

		for (; *s && *s != ' ' && *s != '='; s++)
			key.push_back(*s);
		/* Skip the = */
		if (*s) s++;
			
		/* Skip the quote */
		if (*s) s++;
		for (; *s && *s != '\'' && *s != '\"'; s++)
			value.push_back(*s);
		/* Skip the quote */
		if (*s) s++;
		
		/* Skip the spaces. */
		for (; *s && (*s == ' ' || *s == '\n' || *s == '\r'); s++);

		/* Allow an empty value as it's valid. */
		if (!key.empty()) params[key] = value;
		
		debugprint(debugmode, "XML Parse: Key: %s Value: %s", key.c_str(), value.c_str());
	}
	
	/* Look for the closing flag "/>" at the end of a tag. This means there is no enclosing
	 * tags. */
	if (*s == '/')
	{
		closing = true;
		debugprint(debugmode, "XML Parse: Closing tag");
	}
	
	s++;
	
	return s;
}

/*!
 * \brief Truncate a string to the first /, if any.
 *
 * Probably better written using std::string methods.
 *
 * \param in The string to strip. */
void stripslash(std::string &in)
{
	std::string result;
	
	for (char *s = (char *) in.c_str(); *s && *s != '/'; s++)
		result.push_back(*s);
		
	in = result;
}

/*!
 * \brief Remove all newlines from a string.
 *
 * Ditto about using std::string functions.
 *
 * \param in The string to strip.
 */
void removenewlines(std::string &in)
{
	std::string result;
	
	for (char *s = (char *) in.c_str(); *s; s++)
	{
		if (*s != '\r' && *s != '\n')
			result.push_back(*s);
	}
	
	in = result;
}
